#include "arch/MMU.h"
#include "arch/x86_64/CPU.h"
#include <luna/CString.h>
#include <luna/Check.h>
#include <luna/Stack.h>
#include <luna/Types.h>

struct [[gnu::packed]] GDTR
{
    u16 size;
    u64 offset;
};

static_assert(sizeof(GDTR) == 10UL);

struct [[gnu::packed]] GDTEntry
{
    u16 limit0;
    u16 base0;
    u8 base1;
    u8 access;
    u8 limit1_flags;
    u8 base2;
};

static_assert(sizeof(GDTEntry) == 8UL);

struct [[gnu::packed]] HighGDTEntry
{
    u32 base_high;
    u32 reserved;
};

static_assert(sizeof(HighGDTEntry) == 8UL);

struct [[gnu::packed]] alignas(4096) GlobalDescriptorTable
{
    GDTEntry null;
    GDTEntry kernel_code;
    GDTEntry kernel_data;
    GDTEntry user_code;
    GDTEntry user_data;
    GDTEntry tss;
    HighGDTEntry tss2;
};

TSS task_state_segment;

static GlobalDescriptorTable gdt = { { 0x0000, 0x0000, 0x00, 0x00, 0x00, 0x00 },
                                     { 0xffff, 0x0000, 0x00, 0x9a, 0xaf, 0x00 },
                                     { 0xffff, 0x0000, 0x00, 0x92, 0xcf, 0x00 },
                                     { 0xffff, 0x0000, 0x00, 0xfa, 0xaf, 0x00 },
                                     { 0xffff, 0x0000, 0x00, 0xf2, 0xcf, 0x00 },
                                     { 0x0000, 0x0000, 0x00, 0xe9, 0x0f, 0x00 },
                                     { 0x00000000, 0x00000000 } };

extern "C" void load_gdt(GDTR* gdtr);
extern "C" void load_tr(int segment);

static void set_base(GDTEntry* entry, u32 base)
{
    entry->base0 = (base & 0xFFFF);
    entry->base1 = (base >> 16) & 0xFF;
    entry->base2 = (u8)((base >> 24) & 0xFF);
}

static void set_limit(GDTEntry* entry, u32 limit)
{
    expect(limit <= 0xFFFFF, "Limit too big for a GDT entry");
    entry->limit0 = limit & 0xFFFF;
    entry->limit1_flags = (entry->limit1_flags & 0xF0) | ((limit >> 16) & 0xF);
}

static void set_tss_base(GDTEntry* tss1, HighGDTEntry* tss2, u64 addr)
{
    set_base(tss1, addr & 0xffffffff);
    tss2->base_high = (u32)(addr >> 32);
}

static u8 alternate_stack[ARCH_PAGE_SIZE * 4];

static void setup_tss()
{
    memset(&task_state_segment, 0, sizeof(TSS));
    task_state_segment.iomap_base = sizeof(TSS);
    task_state_segment.ist[0] = Stack { (u64)alternate_stack, ARCH_PAGE_SIZE * 4 }.top();
    set_tss_base(&gdt.tss, &gdt.tss2, (u64)&task_state_segment);
    set_limit(&gdt.tss, sizeof(TSS) - 1);
}

void setup_gdt()
{
    static GDTR gdtr;
    gdtr.offset = (u64)&gdt;
    gdtr.size = sizeof(GlobalDescriptorTable);
    setup_tss();
    load_gdt(&gdtr);
    load_tr(0x2b);
}
